"""WeatherClient main client module."""
from __future__ import annotations

from typing import TYPE_CHECKING

from weather_client.providers.open_meteo_provider import OpenMeteoProvider
from weather_client.providers.owm_provider import OWMProvider

if TYPE_CHECKING:
    from weather_client.models import WeatherReading, WeatherStation


class WeatherClient:
    def __init__(self, latitude: float, longitude: float, owm_api_key: str | None = None):
        """Initialize the WeatherClient.

        Args:
            latitude: Latitude of the location to fetch weather for.
            longitude: Longitude of the location to fetch weather for.
            owm_api_key: Optional OpenWeatherMap API key for enhanced data.
        """
        self.latitude = latitude
        self.longitude = longitude
        self._owm = OWMProvider(owm_api_key) if owm_api_key else None
        self._open_meteo = OpenMeteoProvider()

    def get_weather(
        self,
    ) -> tuple[WeatherReading, list[WeatherReading], WeatherStation]:
        """Fetch weather data from providers, falling back as needed.

        Returns:
            A tuple of (current reading, list of hourly readings, weather station info).
        """
        if self._owm:
            try:
                return self._owm.fetch(self.latitude, self.longitude)
            except Exception:  # noqa: BLE001, S110
                # Provider failures should fall back to Open-Meteo.
                # This includes cases where OWM One Call is not available for the key/tier.
                pass
        return self._open_meteo.fetch(self.latitude, self.longitude)

    def get_open_meteo_weather(
        self,
    ) -> tuple[WeatherReading, list[WeatherReading], WeatherStation]:
        """Fetch weather data from Open Meteo.

        Returns:
            A tuple of (current reading, list of hourly readings, weather station info).
        """
        return self._open_meteo.fetch(self.latitude, self.longitude)



from __future__ import annotations

from datetime import UTC, datetime

from weather_client import WeatherClient
from weather_client.models import WeatherReading, WeatherStation, Wind
from weather_client.providers.open_meteo_provider import OpenMeteoProvider
from weather_client.providers.owm_provider import OWMProvider


def test_client_falls_back_to_open_meteo(mocker):
    utc_time = datetime(2025, 12, 20, 12, 0, tzinfo=UTC)
    expected = (
        WeatherReading(
            utc_time=utc_time,
            local_time=utc_time.astimezone(),
            temperature=20.0,
            sky="clear sky",
            wind=Wind(speed=3.0, deg=180.0),
        ),
        [],
        WeatherStation(source="Open-Meteo", latitude=-33.86, longitude=151.21),
    )

    mocker.patch.object(OWMProvider, "fetch", side_effect=RuntimeError("boom"))
    open_meteo_fetch = mocker.patch.object(OpenMeteoProvider, "fetch", return_value=expected)

    client = WeatherClient(-33.86, 151.21, owm_api_key="dummy")
    assert client.get_weather() == expected
    open_meteo_fetch.assert_called_once()


def test_client_uses_owm_when_available(mocker):
    utc_time = datetime(2025, 12, 20, 12, 0, tzinfo=UTC)

    expected = (
        WeatherReading(
            utc_time=utc_time,
            local_time=utc_time.astimezone(),
            temperature=22.0,
            sky="sunny",
            wind=Wind(speed=2.0, deg=None),
        ),
        [],
        WeatherStation(source="OpenWeatherMap One Call API", latitude=-33.86, longitude=151.21),
    )

    owm_fetch = mocker.patch.object(OWMProvider, "fetch", return_value=expected)
    mocker.patch.object(OpenMeteoProvider, "fetch", side_effect=AssertionError("should not call"))

    client = WeatherClient(-33.86, 151.21, owm_api_key="dummy")
    assert client.get_weather() == expected
    owm_fetch.assert_called_once()



"""WeatherClient provider for OpenWeatherMap (OWM) API."""
from __future__ import annotations

from datetime import UTC, datetime
from typing import Any

import requests
from pyowm import OWM
from pyowm.commons.exceptions import APIRequestError, UnauthorizedError

from weather_client.models import WeatherReading, WeatherStation, Wind


class OWMProvider:
    def __init__(self, api_key: str):
        self._api_key = api_key
        self._owm = OWM(api_key)
        self._mgr = self._owm.weather_manager()

    def fetch(
        self, lat: float, lon: float
    ) -> tuple[WeatherReading, list[WeatherReading], WeatherStation]:
        try:
            one_call = self._mgr.one_call(lat=lat, lon=lon, units="celsius")
        except UnauthorizedError:
            # Many "free" OpenWeatherMap keys don't have access to One Call.
            # Fall back to free-tier endpoints (current + 5-day/3-hour forecast).
            return self._fetch_free(lat, lon)
        except APIRequestError as e:
            error_msg = f"OpenWeatherMap request failed: {e}"
            raise RuntimeError(error_msg) from e

        utc_now = datetime.now(UTC)
        current = one_call.current

        current_reading = WeatherReading(
            utc_time=utc_now,
            local_time=utc_now.astimezone(),
            temperature=current.temperature("celsius")["temp"],
            sky=current.detailed_status,
            wind=Wind(speed=self._covert_wind_speed(current.wind()["speed"]), deg=current.wind().get("deg")),
        )

        hourly: list[WeatherReading] = []
        for h in one_call.forecast_hourly or []:
            ref = h.reference_time("unix")
            if isinstance(ref, datetime):
                utc_ts = ref if ref.tzinfo is not None else ref.replace(tzinfo=UTC)
            else:
                utc_ts = datetime.fromtimestamp(int(ref), tz=UTC)
            if utc_ts.date() != utc_now.date() or utc_ts < utc_now:
                continue
            hourly.append(
                WeatherReading(
                    utc_time=utc_ts,
                    local_time=utc_ts.astimezone(),
                    temperature=h.temperature("celsius")["temp"],
                    sky=h.detailed_status,
                    wind=Wind(speed=self._covert_wind_speed(h.wind()["speed"]), deg=h.wind().get("deg")),
                )
            )

        station = WeatherStation("OpenWeatherMap One Call API", lat, lon)
        return current_reading, hourly, station

    def _fetch_free(  # noqa: PLR0914
        self, lat: float, lon: float
    ) -> tuple[WeatherReading, list[WeatherReading], WeatherStation]:
        """Use free-tier endpoints: current weather + 5 day / 3 hour forecast.

        Args:
            lat (float): Latitude of the location.
            lon (float): Longitude of the location.

        Returns:
            tuple[WeatherReading, list[WeatherReading], WeatherStation]: Current reading, hourly forecast, and station info.
        """

        def _as_float(value: Any, *, field: str) -> float:
            try:
                return float(value)
            except (TypeError, ValueError) as e:
                error_msg = f"OpenWeatherMap response missing/invalid {field}"
                raise RuntimeError(error_msg) from e

        current_url = "https://api.openweathermap.org/data/2.5/weather"
        forecast_url = "https://api.openweathermap.org/data/2.5/forecast"
        params = {"lat": lat, "lon": lon, "appid": self._api_key, "units": "metric"}

        current_resp = requests.get(current_url, params=params, timeout=10)
        current_resp.raise_for_status()
        current_data: dict[str, Any] = current_resp.json()

        utc_now = datetime.now(UTC)
        dt_raw = current_data.get("dt")
        dt_unix = int(dt_raw) if isinstance(dt_raw, (int, float)) else int(utc_now.timestamp())
        utc_current_time = datetime.fromtimestamp(dt_unix, tz=UTC)
        current_weather = (current_data.get("weather") or [{}])[0] or {}
        current_wind = current_data.get("wind") or {}
        # Convert the wind speed from m/s to km/h
        if "speed" in current_wind:
            current_wind["speed"] = self._covert_wind_speed(_as_float(current_wind["speed"], field="wind.speed"))
        current_main = current_data.get("main") or {}

        current_reading = WeatherReading(
            utc_time=utc_current_time,
            local_time=utc_current_time.astimezone(),
            temperature=_as_float(current_main.get("temp"), field="main.temp"),
            sky=str(current_weather.get("description") or "unknown"),
            wind=Wind(
                speed=_as_float(current_wind.get("speed", 0.0), field="wind.speed"),
                deg=(float(current_wind["deg"]) if "deg" in current_wind else None),
            ),
        )

        forecast_resp = requests.get(forecast_url, params=params, timeout=10)
        forecast_resp.raise_for_status()
        forecast_data: dict[str, Any] = forecast_resp.json()

        hourly: list[WeatherReading] = []
        for item in forecast_data.get("list", []):
            utc_ts = datetime.fromtimestamp(int(item.get("dt", 0)), tz=UTC)
            if utc_ts.date() != utc_now.date() or utc_ts < utc_now:
                continue

            weather = (item.get("weather") or [{}])[0] or {}
            wind = item.get("wind") or {}
            main = item.get("main") or {}

            hourly.append(
                WeatherReading(
                    utc_time=utc_ts,
                    local_time=utc_ts.astimezone(),
                    temperature=_as_float(main.get("temp"), field="forecast.main.temp"),
                    sky=str(weather.get("description") or "unknown"),
                    wind=Wind(
                        speed=_as_float(wind.get("speed", 0.0), field="forecast.wind.speed"),
                        deg=(float(wind["deg"]) if "deg" in wind else None),
                    ),
                )
            )

        station = WeatherStation("OpenWeatherMap (free)", lat, lon)
        return current_reading, hourly, station

    @staticmethod
    def _covert_wind_speed(wind: float) -> float:
        """Convert wind speed from m/s to km/h.

        Args:
            wind (float): Wind speed in meters per second.

        Returns:
            float: Wind speed in kilometers per hour.
        """
        return wind * 3.6



from __future__ import annotations

from datetime import UTC, datetime

import pytest
from pyowm.commons.exceptions import UnauthorizedError

from weather_client.providers.owm_provider import OWMProvider


class _FakeResponse:
    def __init__(self, payload, *, status_code: int = 200):
        self._payload = payload
        self.status_code = status_code

    def raise_for_status(self) -> None:
        if self.status_code >= 400:
            error_msg = f"HTTP {self.status_code} Error"
            raise RuntimeError(error_msg)

    def json(self):
        return self._payload


def test_owm_provider_falls_back_to_free_endpoints_on_unauthorized(mocker):
    provider = OWMProvider("dummy")
    mocker.patch.object(
        provider._mgr,  # noqa: SLF001
        "one_call",
        side_effect=UnauthorizedError("Invalid API Key provided"),
    )

    now = datetime(2025, 12, 21, 4, 0, tzinfo=UTC)
    mocker.patch("weather_client.providers.owm_provider.datetime", wraps=datetime)
    mocker.patch("weather_client.providers.owm_provider.datetime.now", return_value=now)

    current_payload = {
        "dt": int(now.timestamp()),
        "main": {"temp": 20.5},
        "weather": [{"description": "clear sky"}],
        "wind": {"speed": 3.2, "deg": 180},
    }

    forecast_payload = {
        "list": [
            {
                "dt": int(now.timestamp()) + 3 * 3600,
                "main": {"temp": 21.0},
                "weather": [{"description": "few clouds"}],
                "wind": {"speed": 2.8, "deg": 170},
            }
        ]
    }

    get = mocker.patch(
        "weather_client.providers.owm_provider.requests.get",
        side_effect=[
            _FakeResponse(current_payload),
            _FakeResponse(forecast_payload),
        ],
    )

    current, hourly, station = provider.fetch(-33.86, 151.21)

    assert station.source == "OpenWeatherMap (free)"
    assert current.temperature == 20.5
    assert hourly
    assert hourly[0].temperature == 21.0
    assert get.call_count == 2


def test_owm_provider_free_endpoint_http_error_raises(mocker):
    provider = OWMProvider("dummy")
    mocker.patch.object(provider._mgr, "one_call", side_effect=UnauthorizedError("nope"))  # noqa: SLF001
    mocker.patch(
        "weather_client.providers.owm_provider.requests.get",
        return_value=_FakeResponse({}, status_code=401),
    )

    with pytest.raises(RuntimeError):
        provider.fetch(-33.86, 151.21)



"""WeatherClient provider for Open Meteo weather provider."""
from __future__ import annotations

from datetime import UTC, datetime
from typing import Any

import requests

from weather_client.models import WeatherReading, WeatherStation, Wind

WEATHER_CODE_MAP = {
    0: "clear sky",
    1: "mainly clear",
    2: "partly cloudy",
    3: "overcast",
    45: "fog",
    61: "rain",
    71: "snow",
    95: "thunderstorm",
}


class OpenMeteoProvider:
    def fetch(
        self, lat: float, lon: float
    ) -> tuple[WeatherReading, list[WeatherReading], WeatherStation]:
        url = (
            "https://api.open-meteo.com/v1/forecast"
            f"?latitude={lat}&longitude={lon}"
            "&current_weather=true"
            "&hourly=temperature_2m,weathercode,windspeed_10m,winddirection_10m"
            "&timezone=UTC"
        )
        response = requests.get(url, timeout=10)
        response.raise_for_status()
        data: dict[str, Any] = response.json()
        utc_now = datetime.now(UTC)

        c = data["current_weather"]
        current_time_raw = c.get("time")
        utc_time = (
            datetime.fromisoformat(current_time_raw).replace(tzinfo=UTC)
            if isinstance(current_time_raw, str)
            else utc_now
        )
        local_time = utc_time.astimezone()
        current_reading = WeatherReading(
            utc_time=utc_time,
            local_time=local_time,
            temperature=c["temperature"],
            sky=WEATHER_CODE_MAP.get(c["weathercode"], "unknown"),
            wind=Wind(speed=self._covert_wind_speed(c["windspeed"]), deg=c.get("winddirection")),
        )

        hourly = []
        for i, ts in enumerate(data["hourly"]["time"]):
            utc_ts_dt = datetime.fromisoformat(ts).replace(tzinfo=UTC)
            local_ts_dt = utc_ts_dt.astimezone()
            if utc_ts_dt.date() != utc_time.date() or utc_ts_dt < utc_time:
                continue
            hourly.append(
                WeatherReading(
                    utc_time=utc_ts_dt,
                    local_time=local_ts_dt,
                    temperature=data["hourly"]["temperature_2m"][i],
                    sky=WEATHER_CODE_MAP.get(data["hourly"]["weathercode"][i], "unknown"),
                    wind=Wind(
                        speed=self._covert_wind_speed(data["hourly"]["windspeed_10m"][i]),
                        deg=data["hourly"]["winddirection_10m"][i],
                    ),
                )
            )

        station = WeatherStation("Open-Meteo", lat, lon)
        return current_reading, hourly, station

    @staticmethod
    def _covert_wind_speed(wind: float) -> float:
        """Convert wind speed from m/s to km/h.

        Args:
            wind (float): Wind speed as provided by Open-Meteo.

        Returns:
            float: Wind speed in kilometers per hour.
        """
        return wind



"""Data models for WeatherClient."""
from dataclasses import dataclass
from datetime import datetime


@dataclass
class Wind:
    speed: float
    deg: float | None
    units: str = "km/h"


@dataclass
class WeatherReading:
    utc_time: datetime
    local_time: datetime
    temperature: float
    sky: str
    wind: Wind


@dataclass
class WeatherStation:
    source: str
    latitude: float
    longitude: float
